#define INCL_DOS
#define INCL_DOSERRORS
#define INCL_LONGLONG
#include <os2.h>

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <process.h>

enum {
  TICKER_STATUS_NONE = 0,
  TICKER_STATUS_INIT,
  TICKER_STATUS_IN_PROGRESS,
  TICKER_STATUS_EXIT
};

enum {
  TICKER_FLAG_DEFAULT = 0,
  TICKER_FLAG_TIMECRITICAL
};

typedef struct _tickerparam {
  int status;
  HEV hev;
  int interval;
  int flag;
  int loopcnt;
  int loopcnt_progress;
  uint64_t *ticklap;
} TickerParam;


static int do_as_timecritical = 0;


APIRET
myDosGetThreadPriority(PTIB ptib, ULONG *ulClass, LONG *lDelta)
{
  APIRET rc = 0;
  PPIB ppib;
  
  if (!ptib)
    rc = DosGetInfoBlocks(&ptib, &ppib);
  if (rc == 0) {
    ULONG t_pri;
    t_pri = ptib->tib_ptib2->tib2_ulpri;
    if (ulClass)
      *ulClass = (t_pri >> 8) & 0xffU;
    if (lDelta) {
      *lDelta = (t_pri & 0x1f); /* negative value will be never set */
    }
  }
  return rc;
}


APIRET
myDosSleep(ULONG msec)
{
  APIRET rc;
  
  if (do_as_timecritical) {
    ULONG ulClass;
    LONG lDelta;
    myDosGetThreadPriority(NULL, &ulClass, &lDelta);
    DosSetPriority(PRTYS_THREAD, PRTYC_TIMECRITICAL, 31L, 0);
    rc = DosSleep(msec);
    DosSetPriority(PRTYS_THREAD, ulClass, lDelta, 0);
  }
  else
    rc = DosSleep(msec);
  
  return rc;
}


APIRET
myDosWaitEventSem(HEV hev, ULONG timeout)
{
  APIRET rc;
  
#if 0
  if (do_as_timecritical) {
    ULONG ulClass;
    LONG lDelta;
    myDosGetThreadPriority(NULL, &ulClass, &lDelta);
    DosSetPriority(PRTYS_THREAD, PRTYC_TIMECRITICAL, 31L, 0);
    rc = DosWaitEventSem(hev, timeout);
    DosSetPriority(PRTYS_THREAD, ulClass, lDelta, 0);
  }
  else
#endif
    rc = DosWaitEventSem(hev, timeout);
  
  return rc;
}

APIRET
myDosAsyncTimer(ULONG msec, HSEM hsem, PHTIMER phtimer)
{
  APIRET rc;
  
  if (do_as_timecritical) {
    ULONG ulClass;
    LONG lDelta;
    myDosGetThreadPriority(NULL, &ulClass, &lDelta);
    DosSetPriority(PRTYS_THREAD, PRTYC_TIMECRITICAL, 31L, 0);
    rc = DosAsyncTimer(msec, hsem, phtimer);
    DosSetPriority(PRTYS_THREAD, ulClass, lDelta, 0);
  }
  else
    rc = DosAsyncTimer(msec, hsem, phtimer);
  
  return rc;
}




static
void sleeploop_thread(void *param)
{
  TickerParam *p = (TickerParam *)param;
  uint64_t *t;
  int i;
  
  p->status = TICKER_STATUS_INIT;
  t = p->ticklap;
  p->status = TICKER_STATUS_IN_PROGRESS;
  
  for(i=0; i< p->loopcnt; i++ /* atomic-increment is needed? */) {
    p->loopcnt_progress = i;
    myDosSleep(p->interval);
    if (p->hev) DosPostEventSem(p->hev);
    DosTmrQueryTime((PQWORD)(t + i));
  }
  
  p->status = TICKER_STATUS_EXIT;
  if (p->hev) DosPostEventSem(p->hev);
  
  _endthread();
}


void
print_ticklaps(TickerParam *p)
{
  int i;
  ULONG ul;
  uint64_t t, t0, td;
  
  DosTmrQueryFreq(&ul);
  for(i=0; i<p->loopcnt; i++) {
    t = p->ticklap[i];
    printf("lap%4d %llu", i, (t * 1000ULL)/ul);
    if (i > 0) {
      t0 = p->ticklap[i-1];
      td = t - t0;
      printf(" (%llu ms)", (td * 1000ULL)/ul);
    }
    printf("\n");
  }
}


void
CountSleepThread(TickerParam *p)
{
  APIRET rc;
  int tid;
  ULONG ul;

  tid = _beginthread(sleeploop_thread, NULL, 65536, p);
  
  while(1) {
    rc = DosWaitEventSem(p->hev, SEM_INDEFINITE_WAIT);
    if (rc != 0 && rc != ERROR_INTERRUPT) {
      printf("err:DosWaitEventSem (rc=%d)\n", rc);
      break;
    }
    DosResetEventSem(p->hev, &ul);
    if (p->status == TICKER_STATUS_EXIT) break;
  }
  return ;
}

void
CountAsyncTimer(TickerParam *p)
{
  APIRET rc, rc2;
  HTIMER htm;
  uint64_t *t;
  int i;
  ULONG ul;
  
  p->status = TICKER_STATUS_INIT;
  t = p->ticklap;
  p->status = TICKER_STATUS_IN_PROGRESS;
  
  for(i=0; i< p->loopcnt; i++) {
    htm = NULLHANDLE;
    p->loopcnt_progress = i;
    rc = myDosAsyncTimer(p->interval, (HSEM)(p->hev), &htm);
    if (rc) {
      printf("err:DosAyncTimer (rc=%d)\n", rc);
      return;
    }
    rc = DosWaitEventSem(p->hev, SEM_INDEFINITE_WAIT);
    if (rc != 0 && rc != ERROR_INTERRUPT) {
      printf("err:DosWaitEventSem (rc=%d)\n", rc);
      break;
    }
    DosTmrQueryTime((PQWORD)(t + i));
    DosResetEventSem(p->hev, &ul);
    /* DosStopTimer(htm) is not needed (rc=ERROR_TS_HANDLE). */;
  }
  
  p->status = TICKER_STATUS_EXIT;
}


static int method_thread = 0;
static int method_async = 0;
static int opt_h = 0;
static int opt_t = 0;
static int opt_n = 10;

void
usage()
{
  const char msg[] = \
    "tsttick [-n(count)] [-t] [thread] [async]\n"
    "  -n       loop count (default: -n10)\n"
    "  -t       do as \"time-critical\" job\n"
    " thread    method: sleep-loops in the child thread.\n"
    " async     method: DosAsyncTimer\n";
  
  printf("%s\n", msg);
}

int
getopt(int argc, char **argv)
{
  char c, *s;
  
  while(argc) {
    s = *argv;
    c = *s;
    if (c == '-' || c == '/') {
      switch(s[1]) {
        case 'H': case 'h': case '?':
          opt_h = 1;
          break;
        case 'T': case 't':
          opt_t = 1;
          break;
        case 'N': case 'n':
          opt_n = atoi(s + 2);
          break;
      }
    }
    else {
      if (stricmp(s, "thread") == 0)
        method_thread = 1;
      if (stricmp(s, "async") == 0)
        method_async = 1;
    }
    --argc;
    ++argv;
  }
}


int
main(int argc, char **argv)
{
#define TICK_LIMIT 1000
  TickerParam tp;
  APIRET rc;
  uint64_t tick[TICK_LIMIT];
  ULONG ul;
  
  getopt(argc-1, argv+1);
  if (opt_h || !(method_thread || method_async)) {
    usage();
    return 0;
  }
  if (opt_n <= 0 || opt_n > TICK_LIMIT) {
    fprintf(stderr, "parameter err.\n");
    exit(1);
  }
  do_as_timecritical = opt_t;
  
  memset(&tp, 0, sizeof(tp));
  rc = DosCreateEventSem((PSZ)NULL, &tp.hev, DC_SEM_SHARED, 0);
  if (rc) {
    printf("err:DosCreateEventSem (rc=%d)\n", rc);
    exit(1);
  }
  tp.interval = 1;
  tp.ticklap = tick;
  tp.loopcnt = opt_n;
  
  if (method_thread) {
    printf("Method: thread\n");
    CountSleepThread(&tp);
    print_ticklaps(&tp);
  }
  if (method_async) {
    printf("Method: DosAsyncTimer\n");
    CountAsyncTimer(&tp);
    print_ticklaps(&tp);
  }
  
  return 0;
}

